package vsp;

import vsp.gaussian.Params;
import vsp.gaussian.RandomGenerator;
import cz.zcu.fav.kiv.jsim.JSimInvalidParametersException;
import cz.zcu.fav.kiv.jsim.JSimLink;
import cz.zcu.fav.kiv.jsim.JSimProcess;
import cz.zcu.fav.kiv.jsim.JSimSecurityException;
import cz.zcu.fav.kiv.jsim.JSimSimulation;
import cz.zcu.fav.kiv.jsim.JSimSimulationAlreadyTerminatedException;
import cz.zcu.fav.kiv.jsim.JSimSystem;
import cz.zcu.fav.kiv.jsim.JSimTooManyProcessesException;

/**
 * Transaction generator
 * 
 * @author Jan Tichava <jtichava@students.zcu.cz>
 *
 */
public class Generator extends JSimProcess {
	
	/**
	 * Silent mode for generating Gaussian
	 */
	private static final boolean SILENT = true;
	
	/**
	 * Exponential distribution ratio
	 */
	private double lambda;
	/**
	 * Queue for server
	 */
	private QueueWithServer queue;

	/**
	 * Number of generated transactions
	 */
	private int generated;
	
	/**
	 * Median for Gaussian
	 */
	private double median;
	/**
	 * Deviation for Gaussian
	 */
	private double deviation;
	
	private boolean gauss;
	
	/**
	 * Creates exponential new generator
	 * 
	 * @param name name of generator
	 * @param simulation simulation object
	 * @param lambda lambda
	 * @param queue output queue
	 * @throws JSimSimulationAlreadyTerminatedException
	 * @throws JSimInvalidParametersException
	 * @throws JSimTooManyProcessesException
	 */
	public Generator(String name, JSimSimulation simulation, double lambda, QueueWithServer queue) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException {
		super(name, simulation);
		
		this.lambda = lambda;
		this.queue = queue;
		
		generated = 0;
		gauss = false;
	}
	
	/**
	 * 
	 * Creates new Gaussian generator
	 * 
	 * @param name name of generator
	 * @param simulation simulation object
	 * @param median median
	 * @param deviation deviation
	 * @param queue output queue 
	 * @throws JSimSimulationAlreadyTerminatedException
	 * @throws JSimInvalidParametersException
	 * @throws JSimTooManyProcessesException
	 */
	public Generator(String name, JSimSimulation simulation, double median, double deviation, QueueWithServer queue) throws JSimSimulationAlreadyTerminatedException, JSimInvalidParametersException, JSimTooManyProcessesException {
		super(name, simulation);
		
		this.median = median;
		this.deviation = deviation;
		this.queue = queue;
		
		generated = 0;
		gauss = true;
	}
	
	/**
	 * In infinite loop:
	 * - generates new transaction
	 * - generates random number from <0; 1>
	 * - adds transation to queue
	 * - pause process in depend of lambda 
	 */
	protected void life() {
		try {
			
			while (true) {
				JSimLink link = new JSimLink(new Transaction(myParent.getCurrentTime()));
				
				queue = addTransaction(link, queue);

				generated++;
				
				double workTime = 0;
				if (gauss) {
					Params params = new Params(1, median, deviation);
					RandomGenerator generator = new RandomGenerator(params);
					workTime = Math.abs(generator.generate(SILENT)[0]);
				}
				else {
					workTime = JSimSystem.negExp(lambda);
				}

				hold(workTime);
			}
			
		} catch (JSimSecurityException e) {
			e.printStackTrace();
		} catch (JSimInvalidParametersException e) {
			e.printStackTrace();
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/**
	 * Add transaction to queue
	 * 
	 * @param link
	 * @param queue
	 * @return
	 * @throws JSimSecurityException
	 * @throws JSimInvalidParametersException
	 */
	private QueueWithServer addTransaction(JSimLink link, QueueWithServer queue) throws JSimSecurityException, JSimInvalidParametersException {
		
		link.into(queue);
		if (queue.getServer().isIdle()) {
			queue.getServer().activate(myParent.getCurrentTime());
		}
		
		return queue;
	}

	/**
	 * Get number of generated transaction
	 * 
	 * @return the generated
	 */
	public int getGenerated() {
		return generated;
	}
	
	/**
	 * Get median requests count
	 * @return
	 */
	public double getLq() {
		return myParent.getCurrentTime() / getGenerated();
	}
	
	
}
