package vsp;

import java.io.IOException;
import java.util.ArrayList;

import vsp.config.Config;
import vsp.config.ConfigExp;
import vsp.config.ConfigGauss;
import vsp.readConfig.ReadConfigExp;
import vsp.readConfig.ReadConfigGauss;
import cz.zcu.fav.kiv.jsim.JSimInvalidParametersException;
import cz.zcu.fav.kiv.jsim.JSimMethodNotSupportedException;
import cz.zcu.fav.kiv.jsim.JSimSecurityException;
import cz.zcu.fav.kiv.jsim.JSimSimulation;
import cz.zcu.fav.kiv.jsim.JSimSimulationAlreadyTerminatedException;
import cz.zcu.fav.kiv.jsim.JSimTooManyHeadsException;
import cz.zcu.fav.kiv.jsim.JSimTooManyProcessesException;

/**
 * 
 * VSP
 * 
 * @author Jan Tichava <jtichava@students.zcu.cz>
 *
 */
public class Vsp {
	
	/**
	 * When is only output queue 
	 */
	private static final QueueWithServer NO_QUEUE = null;
	/**
	 * When is only output queue
	 */
	private static final double NO_PROBABILITY = 2;
	
	/**
	 * Default variation for default Gaussian
	 */
	private static final double VARIATION_1 = 0.05;
	/**
	 * Default variation for default Gaussian
	 */
	private static final double VARIATION_2 = 0.2;
	/**
	 * Default variation for default Gaussian
	 */
	private static final double VARIATION_3 = 0.7;

	
	
	/**
	 * Main simulation
	 * @param args command line arguments
	 * @throws IOException 
	 * @throws NumberFormatException 
	 */
	public Vsp(String[] args) throws NumberFormatException, IOException {
		CommandLine commandLine = parseArguments(args);
		
		if (commandLine.getFileName() == null) {
			if (commandLine.getGauss()) {
				runSimulation(new ConfigGauss(VARIATION_1, commandLine.getSteps()));
				runSimulation(new ConfigGauss(VARIATION_2, commandLine.getSteps()));
				runSimulation(new ConfigGauss(VARIATION_3, commandLine.getSteps()));
			}
			else {
				runSimulation(new ConfigExp(commandLine.getSteps()));
			}
		}
		else {
			if (commandLine.getGauss()) {
				ReadConfigGauss readConfigGauss = new ReadConfigGauss(commandLine.getFileName());
				runSimulation(readConfigGauss.read());
			}
			else {
				ReadConfigExp readConfigExp = new ReadConfigExp(commandLine.getFileName());
				runSimulation(readConfigExp.read());
			}
		}
	}
	
	/**
	 * Parse command line arguments
	 * @param args command line argument
	 * @return number of steps
	 */
	private CommandLine parseArguments(String[] args) {
		CommandLine commandLine = new CommandLine();

		try {
			switch (args.length) {
			case 0:
				printHelp();
				break;
			case 3:
				commandLine.setFileName(args[2]);
			case 2:
				commandLine.setSteps(args[0]);
				commandLine.setGauss(args[1]);
				break;
				
			default:
				System.out.println("Too much or too less arguments!");
				printHelp();
				System.exit(0);
				break;
			}
		} catch (NumberFormatException e) {
			System.out.println("Error: " + e.getMessage());
			System.exit(0);
		} catch (Exception e) {
			System.out.println("Error: " + e.getMessage());
			System.exit(0);
		}

		return commandLine;
	}
	
	/**
	 * Print help
	 */
	private void printHelp() {
		System.out.println("Help:");
		System.out.println("\tjava -jar Vsp.jar count EXP|GAUSS [filename.ext]");
	}
	
	/**
	 * Run simulation
	 * @param config
	 */
	private void runSimulation(Config config) {
		System.out.println("-------------------------------------------");
		
		ArrayList<QueueWithServer> queues = new ArrayList<QueueWithServer>();
		ArrayList<Server> servers = new ArrayList<Server>();
		ArrayList<Generator> generators = new ArrayList<Generator>();
		
		JSimSimulation simulation = null;
		
		try {
			
			System.out.println("Initializing the simulation of an office...");
			
			simulation = new JSimSimulation("Office");
			
			QueueWithServer queueElectronics = new QueueWithServer("Queue electronics request", simulation);
			queues.add(queueElectronics);
			QueueWithServer queuePaper = new QueueWithServer("Queue paper request", simulation);
			queues.add(queuePaper);
			QueueWithServer queueStamp = new QueueWithServer("Queue for stamp", simulation);
			queues.add(queueStamp);
			QueueWithServer queueValidate = new QueueWithServer("Queue for validation", simulation);
			queues.add(queueValidate);
			
			Generator generatorElectronics = null;
			Generator generatorPaper = null;
			
			Server serverElectronics = null;
			Server serverPaper = null;
			Server serverStamp = null;
			Server serverValidate = null;
			
			if (config instanceof ConfigGauss) {
				ConfigGauss configGauss = (ConfigGauss) config;
				
				serverElectronics = new Server("Server electronics request", simulation, configGauss.getMi1(), configGauss.getDeviation1(), queueElectronics, queueStamp, queueElectronics, configGauss.getP1());
				serverPaper = new Server("Server paper request", simulation, configGauss.getMi2(), configGauss.getDeviation2(), queuePaper, queueStamp, queuePaper, configGauss.getP2());
				serverStamp = new Server("Server for stamp", simulation, configGauss.getMi3(), configGauss.getDeviation3(), queueStamp, queueValidate, queueValidate, NO_PROBABILITY);
				serverValidate = new Server("Server for validation", simulation, configGauss.getMi4(), configGauss.getDeviation4(), queueValidate, NO_QUEUE, queuePaper, configGauss.getP3());
				
				generatorElectronics = new Generator("Generator electronics", simulation, configGauss.getMiElectronics(), configGauss.getSigmaElectronics(), queueElectronics);
				generatorPaper = new Generator("Generator paper", simulation, configGauss.getMiPaper(), configGauss.getSigmaPaper(), queuePaper);
			}
			else {
				ConfigExp configExp = (ConfigExp) config;
				
				serverElectronics = new Server("Server electronics request", simulation, configExp.getTs1(), queueElectronics, queueStamp, queueElectronics, configExp.getP1());
				serverPaper = new Server("Server paper request", simulation, configExp.getTs2(), queuePaper, queueStamp, queuePaper, configExp.getP2());
				serverStamp = new Server("Server for stamp", simulation, configExp.getTs3(), queueStamp, queueValidate, queueValidate, NO_PROBABILITY);
				serverValidate = new Server("Server for validation", simulation, configExp.getTs4(), queueValidate, NO_QUEUE, queuePaper, configExp.getP3());
				
				generatorElectronics= new Generator("Generator electronics", simulation, configExp.getLambdaElectronics(), queueElectronics);
				generatorPaper = new Generator("Generator paper", simulation, configExp.getLambdaPaper(), queuePaper);
			}
			
			servers.add(serverElectronics);
			servers.add(serverPaper);
			servers.add(serverStamp);
			servers.add(serverValidate);
			
			generators.add(generatorElectronics);
			generators.add(generatorPaper);
			
			queueElectronics.setServer(serverElectronics);
			queuePaper.setServer(serverPaper);
			queueStamp.setServer(serverStamp);
			queueValidate.setServer(serverValidate);
			
			Statistics statistics = new Statistics(queueStamp, config.getScanTime());
			
			simulation.message("Activating the generators...");
			
			statistics.start();
			generatorElectronics.activate(0);
			generatorPaper.activate(0);
			
			
			simulation.message("Running the simulation, please wait.");
			
			int i = 0;
			while ((i < config.getSteps()) && (simulation.step() == true)) {
				i++;
			}
			
			statistics.stopCollecting();
			
			int total = 0;
			simulation.message("Simulation interrupted at time " + simulation.getCurrentTime());
			for (Generator g : generators) {
				simulation.message(g.getName() + " generated transactions: " + g.getGenerated());
				total += g.getGenerated();
			}
			simulation.message("Queues statistics:");
			for (QueueWithServer q : queues) {
				simulation.message(q.getHeadName() + ": Lw = " + q.getLw());
			}
			
			double numeratora = 0;
			double denominator = 0;
			simulation.message("Servers statistics:");
			for (Server s : servers) {
				simulation.message(s.getName() + ": Tq = " + s.getTq());
				simulation.message(s.getName() + ": load = " + s.getLoad());
				
				numeratora += s.getTransTq();
				denominator += s.getCounter();
			}
			simulation.message("Global statistics:");
			simulation.message("Lq = " + (generatorElectronics.getLq() + generatorPaper.getLq()));
			
			simulation.message("Tq = " + (numeratora/denominator));
			
			simulation.message("Total number requestes: " + total);
			simulation.message("Total number of accepted requestes: " + serverValidate.getCounter());
			simulation.message("Total time of requestes at the office: " + serverValidate.getTransTq());
			simulation.message("Statistics of queue: " + statistics.getQueueName());
			simulation.message("Median: " + statistics.getMedian());
			simulation.message("Variance: " + statistics.getVariance());
			simulation.message("Sigma: " + statistics.getSigma());
			simulation.message("Min items in queue: " + statistics.getMinLength());
			simulation.message("Max items in queue: " + statistics.getMaxLength());
			simulation.message("Histogram:");
			simulation.message(statistics.getHistogramGraphics());
			
			
		} catch (JSimInvalidParametersException e) {
			e.printStackTrace();
		} catch (JSimTooManyHeadsException e) {
			e.printStackTrace();
		} catch (JSimSimulationAlreadyTerminatedException e) {
			e.printStackTrace();
		} catch (JSimTooManyProcessesException e) {
			e.printStackTrace();
		} catch (JSimSecurityException e) {
			e.printStackTrace();
		} catch (JSimMethodNotSupportedException e) {
			e.printStackTrace();
		} finally {
			simulation.shutdown();
		}	
	}
	
	/**
	 * Create main object
	 * @param args command line arguments
	 */
	public static void main(String[] args) {
		try {
			new Vsp(args);
		} catch (NumberFormatException e) {
			System.out.println("Error: " + e.getMessage());
		} catch (IOException e) {
			System.out.println("Error: " + e.getMessage());
		}
	}
}
